"""Base class to create Location objects."""

from __future__ import annotations

from agentpy.objects import Object
from agentpy.sequences import AgentList
import networkx as nx

import pop2net.utils as utils

from . import agent as _agent
from . import model as _model


class Location(Object):
    """Base class for location objects."""

    def __init__(self, model: _model.Model) -> None:
        """Location constructor.

        Args:
            model: Model this Location should be associated with.
        """
        super().__init__(model)
        self.model = model
        self.model.add_location(self)

    @property
    def agents(self) -> AgentList:
        """Return the list of agents affiliated with this location.

        Returns:
            List of agents at this location.
        """
        return self.model.agents_of_location(self)

    def add_agent(self, agent: _agent.Agent) -> None:
        """Assigns the given agent to this location.

        Args:
            agent: The agent that should be added to the location.
        """
        self.model.add_agent_to_location(self, agent)

    def add_agents(self, agents: list) -> None:
        """Add multiple agents at once.

        Args:
            agents (list): An iterable over agents.
        """
        for agent in agents:
            self.add_agent(agent)

    def remove_agent(self, agent: _agent.Agent) -> None:
        """Removes the given agent from this location.

        Args:
            agent: Agent that is to be removed.
        """
        self.model.remove_agent_from_location(
            location=self,
            agent=agent,
        )

    def remove_agents(self, agents: list) -> None:
        """Remove multiple agents at once.

        Args:
            agents (list): An iterable over agents.
        """
        for agent in agents:
            self.remove_agent(agent=agent)

    def neighbors(self, agent: _agent.Agent) -> AgentList:
        """Returns a list of agents which are connected to the given agent via this location.

        Args:
            agent: Agent of whom the neighbors at this location are to be returned.

        Returns:
            AgentList: A list of all agents at this location except the passed agent.
        """
        agents = self.model.agents_of_location(self)
        agents.remove(agent)
        return agents

    def set_weight(self, agent, weight) -> None:
        """Set the weight of an agent at the current location.

        Args:
            agent (Agent): The agent.
            weight (float): The weight.
        """
        self.model.set_weight(agent=agent, location=self, weight=weight)

    def get_weight(self, agent: _agent.Agent) -> float:
        """Return the edge weight between an agent and the location.

        Args:
            agent: Agent of which the edge weight should be returned.

        Returns:
            Edge weight.
        """
        return self.model.get_weight(agent=agent, location=self)

    def project_weights(self, agent1: _agent.Agent, agent2: _agent.Agent) -> float:
        """Calculates the edge weight between two agents assigned to the same location instance.

        Defines how the weights are combined when the edge weight between two agents is determined.
        Can be completely rewritten to have location-specific methods of this kind with the same
        name or can be used as it is in the simulation code.

        Args:
            agent1: First agent of the pair.
            agent2: Second agent of the pair.

        Raises:
            Exception: Raised if `self.weight_projection_function` is not in ["average", "simultan"]

        Returns:
            Combined edge weight.
        """
        return min([self.get_weight(agent1), self.get_weight(agent2)])


class MagicLocation(Location):
    """Helper class to create locations from inside the Creator."""

    n_agents: int | None = None
    overcrowding: bool = None
    only_exact_n_agents: bool = False
    n_locations: int | None = None
    static_weight: bool = False
    recycle: bool = True
    nxgraph: nx.Graph | None = None

    def __init__(self, model: _model.Model) -> None:
        """Create a helper class to create locations.

        Args:
            model (_model.Model):  The model.
        """
        super().__init__(model)
        self.group_id: int | None = None
        self.subgroup_id: int | None = None
        self.group_value: int | str | None = None
        self.subgroup_value: int | str | None = None

    def setup(self) -> None:
        """Use this method to set instance attributes, for instance.

        This method is called automatically by the creator after creating an instance.
        """

    def filter(self, agent: _agent.Agent) -> bool:  # noqa: ARG002
        """Check whether the agent is meant to join this type of location.

        This is a boilerplate implementation of this method which always returns True; i.e. all
        agents will always be allowed at this location. Override this method in your own
        implementations as you seem fit.

        Args:
            agent: The agent that is currently processed by the Creator.

        Returns:
            True if the agent is allowed to join the location, False otherwise.
        """
        return True

    def bridge(self, agent: _agent.Agent) -> float | str | list | None:  # noqa: ARG002
        """Create locations with one agent for each unique value returned.

        Cannot be used in combination with melt().

        Args:
            agent (_agent.Agent): The agent that is currently processed by the Creator.

        Returns:
            float | str | list | None:
                The value which is used to assign agents to location instances.
        """
        return None

    def split(self, agent: _agent.Agent) -> float | str | list | None:  # noqa: ARG002
        """Creates seperate location instances for each unique returned value.

        Args:
            agent: The agent that is currently processed by the Creator.

        Returns:
            float | str | list | None:
                The value(s) that determine(s) to which location instance the agent is assigned.
        """
        return None

    def weight(self, agent: _agent.Agent) -> float | None:  # noqa: ARG002
        """Defines the edge weight between the agent and the location instance.

        Defines how the edge weight between an agent and the location is determined.
        This is a boilerplate implementation of this method which always returns 1; i.e. all
        edge weights will be 1. Override this method in your own implementations as you seem fit.

        Args:
            agent: The agent that is currently processed by the Creator.

        Returns:
            The edge weight.
        """
        return None

    def stick_together(self, agent: _agent.Agent) -> float | str:
        """Assigns agents with a shared value on an attribute to the same location instance.

        Args:
            agent (_agent.Agent): The agent that is currently processed by the Creator.

        Returns:
            float | str: A value that defines the groups of agents.
        """
        return agent.id

    def nest(self) -> type[Location] | None:
        """Nests this location class into another location class.

        Ensures that the agents assigned to the same instance of this location class
        are also assigned to the same instance of the returned location class.

        Returns:
           type[Location] | None: The location class in which this location class should be nested.
        """
        return None

    def melt(self) -> list[Location] | tuple[Location]:
        """Allows to assign agents with different attribute values to the same location.

        Merges the agents assigned to the instances of the returned location classes
        into one instance of this location class.

        Returns:
            None | list | tuple: A list or tuple of location classes.
        """
        return []

    def refine(self):
        """An action that is performed after all location instances have been created."""
        pass

    def _update_weight(self, agent: _agent.Agent) -> None:
        """Create or update the agent-speific weight.

        Args:
            agent: The agent to be updated.
        """
        self.set_weight(agent, self.weight(agent))

    def _update_weights(self) -> None:
        """Update the weight of every agent on this location."""
        for agent_ in self.agents:
            self._update_weight(agent_)

    def _subsplit(self, agent: _agent.Agent) -> str | float | list | None:  # noqa: ARG002
        """Splits a location instance into sub-instances to create a certain network structure.

        Args:
            agent (_agent.Agent): The agent that is currently processed by the Creator.

        Returns:
            str | float | None: A value or a list of values that represent specific sub-instances.
        """
        if self.nxgraph is None:
            return None
        else:
            node_indices = list(self.nxgraph.nodes)
            agent_pos = self.agents_.index(agent)

            if agent_pos <= len(node_indices) - 1:
                node_index = node_indices[agent_pos]

                return [
                    utils._join_positions(pos1=node_index, pos2=neighbor)
                    for neighbor in self.nxgraph.neighbors(node_index)
                ]

            else:
                return None


class MeltLocation(Location):
    """Helper class to melt locations."""

    n_agents: int | None = None
    overcrowding: bool | None = None
    only_exact_n_agents: bool = False
    n_locations: int | None = None

    def filter(self, agent: _agent.Agent) -> bool:  # noqa: ARG002
        """Check whether the agent is meant to join this type of location.

        This is a boilerplate implementation of this method which always returns True; i.e. all
        agents will always be allowed at this location. Override this method in your own
        implementations as you seem fit.

        Args:
            agent: The agent that is currently processed by the Creator.

        Returns:
            True if the agent is allowed to join the location, False otherwise.
        """
        return True

    def stick_together(self, agent: _agent.Agent) -> float | str:
        """Assigns agents with a shared value on an attribute to the same location instance.

        Args:
            agent (_agent.Agent): The agent that is currently processed by the Creator.

        Returns:
            float | str: A value that defines the groups of agents.
        """
        return agent.id

    def split(self, agent: _agent.Agent) -> float | str | list | None:  # noqa: ARG002
        """Creates seperate location instances for each unique returned value.

        Args:
            agent: The agent that is currently processed by the Creator.

        Returns:
            float | str | list | None: The value(s) that determine(s) to which location instance
                the agent is assigned.
        """
        return None

    def weight(self, agent: _agent.Agent) -> float | None:  # noqa: ARG002
        """Defines the edge weight between the agent and the location instance.

        Defines how the edge weight between an agent and the location is determined.
        This is a boilerplate implementation of this method which always returns 1; i.e. all
        edge weights will be 1. Override this method in your own implementations as you seem fit.

        Args:
            agent: The agent that is currently processed by the Creator.

        Returns:
            The edge weight.
        """
        return None



"""Create a population for the simulation."""

from __future__ import annotations

import math
import random
import warnings

import pandas as pd

import pop2net as p2n
import pop2net.utils as utils

from .exceptions import Pop2netException


class Creator:
    """Creates and connects agents and locations."""

    def __init__(
        self,
        model: p2n.Model,
        seed: int = None,
    ) -> None:
        """Instantiate a creator for a specific model.

        Args:
            model (p2n.Model): Model, for which a population should be created
            seed (int, optional): A seed for reproducibility. Defaults to 999.
        """
        self.model = model
        self.seed = seed
        self.rng = random.Random(seed)
        self._dummy_model = p2n.Model()
        self._temp_agent_attrs = ["_P2NTEMP_split_values", "_P2NTEMP_melt_location_weight"]

    def _create_dummy_location(self, location_cls) -> p2n.Location:
        location = location_cls(model=self._dummy_model)
        location.setup()
        return location

    def draw_sample(
        self,
        df: pd.DataFrame,
        n: int | None = None,
        sample_level: str | None = None,
        sample_weight: str | None = None,
        replace_sample_level_column: bool = True,
    ) -> pd.DataFrame:
        """Draw a sample from a dataframe.

        Args:
            df (:class:`pandas.DataFrame`): a pandas DataFrame
            n (int): Target size of the final sample. If this is higher the size of the input
                DataFrame, the sampling will occure with replacement. Without otherwise. If `n` is
                set to `None`, df is returned as it is.
            sample_level (str, optional): A variable the specifies sample units,
                i.e. rows that should always
            be sampled together. For instance, a household ID to sample by households.
                Defaults to None.
            sample_weight (str, optional): The column of df which should be used as
                probability weight. Defaults to None.
            replace_sample_level_column (bool): Should the original values of the sample level be
                overwritten by unique values after sampling to avoid duplicates?

        Returns:
            A pandas DataFrame.
        """
        df = df.copy()

        if n is None:
            return df

        elif sample_level is None:
            df_sample = df.sample(
                n=n,
                replace=not (n <= len(df) and sample_weight is None),
                weights=sample_weight,
                random_state=self.seed,
            )

        else:
            sample_level_ids = list(df[sample_level].drop_duplicates())

            if sample_weight is not None:
                weights = list(df.drop_duplicates(subset=sample_level)[sample_weight])

            samples = []
            counter = 0
            sample_cluster_id = 1
            while counter < n:
                if sample_weight is None:
                    random_id = self.rng.choice(sample_level_ids)
                else:
                    random_id = self.rng.choices(sample_level_ids, weights=weights, k=1)[0]

                sample = df.loc[df[sample_level] == random_id, :].copy()

                # create new unique ids for sample level variable
                if replace_sample_level_column:
                    sample.loc[:, sample_level + "_original"] = sample.loc[:, sample_level]
                    sample.loc[:, sample_level] = sample_cluster_id

                samples.append(sample)

                counter += len(sample)
                sample_cluster_id += 1

            df_sample = pd.concat(samples).reset_index(drop=True)

        return df_sample

    def create_agents(
        self,
        agent_class=p2n.Agent,
        agent_class_attr: None | str = None,
        agent_class_dict: None | dict = None,
        df: pd.DataFrame | None = None,
        n: int | None = None,
        clear: bool = False,
    ) -> p2n.AgentList:
        """Creates agents from a pandas DataFrame.

        Creates one agent-instance of the given agent-class for each row of the given df,
        if df is not None.
        All columns of the df are added as instance attributes containing the row-specific values
        of the specific column.
        If df is None and n is not None, n default agents without any additional attributes are
        created.

        Args:
            agent_class: A class to instantiate all agents with. Every column in the DataFrame will
                result in an attribute of the agents.
            df: The DataFrame from which the agents should be created from.
            n: The number of agents that should be created. Defaults to None.
            clear (bool): Should the agents already included in the model be removed?

        Returns:
            A list of agents.
        """
        if clear:
            self.model.remove_agents(self.model.agents)

        if df is not None:
            df = df.copy()

            # create one agent for each row in df
            agents = []
            for _, row in df.iterrows():
                if agent_class_dict is None:
                    agent = agent_class(model=self.model)
                else:
                    agent = agent_class_dict[row[agent_class_attr]](model=self.model)

                for col_name in df.columns:
                    if col_name == "id":
                        msg = "You are not allowed to set an agent attribute called `id`."
                        raise Exception(msg)
                    else:
                        setattr(agent, col_name, row[col_name])
                agents.append(agent)

        else:
            if n is not None:
                agents = [agent_class(model=self.model) for _ in range(n)]
            else:
                msg = "Either `df` or `n` must be not None."
                raise Exception(msg)

        agents = p2n.AgentList(model=self.model, objs=agents)

        return agents

    def _get_affiliated_agents(self, agents, dummy_location) -> list:
        temp_filter_attr = "_P2NTEMP_filter_" + dummy_location.type

        affiliated_agents = []
        for agent in agents:
            if not hasattr(agent, temp_filter_attr):
                setattr(agent, temp_filter_attr, dummy_location.filter(agent))
                self._temp_agent_attrs.append(temp_filter_attr)

            if getattr(agent, temp_filter_attr):
                affiliated_agents.append(agent)

        return affiliated_agents

    def _get_mother_group_id(self, agent, dummy_location) -> str:
        if dummy_location.nest() is None:
            return "None"

        else:
            # search for mother location assigned to this agent
            n_mother_locations_found = 0
            for location in agent.locations:
                if isinstance(location, dummy_location.nest()):
                    mother_location = location
                    n_mother_locations_found += 1

            # Check if the number of mother locations is not 1
            if n_mother_locations_found > 1:
                warnings.warn(
                    f"""For agent {agent},
                    {n_mother_locations_found} locations of class
                    {dummy_location.nest()} were found as potential mothers of
                    {dummy_location} in the list of locations.""",
                    stacklevel=2,
                )
            elif n_mother_locations_found == 0:
                return "None"

            return "-".join([str(mother_location.split_value), str(mother_location.group_id)])

    def _get_split_values(
        self,
        agents: list,
        dummy_location,
        allow_nesting: bool = False,
    ) -> list[int | str]:
        all_values = []
        for agent in agents:
            agent_values = utils._to_list(dummy_location.split(agent))

            if allow_nesting:
                # Add mother location's value to the value of the lower level location
                for i, value in enumerate(agent_values):
                    mother_group_id = self._get_mother_group_id(agent, dummy_location)
                    agent_values[i] = "-".join([mother_group_id, str(value)])
                    temp_attr = f"_P2NTEMP_{dummy_location.type}_mother_group_id"
                    setattr(agent, temp_attr, mother_group_id)
                    self._temp_agent_attrs.append(temp_attr)

            # Temporarely store group values as agent attribute
            # to assign them to the corresponding location group later
            agent._P2NTEMP_split_values = agent_values

            for value in agent_values:
                if value not in all_values:
                    all_values.append(value)

        return all_values

    def _get_stick_value(self, agent, dummy_location):
        stick_value = dummy_location.stick_together(agent)
        if stick_value is None:
            return "None" + str(agent.id)
        else:
            return stick_value

    def _get_groups(self, agents, location_cls) -> list[list]:
        overcrowding_i = 0

        dummy_location = self._create_dummy_location(location_cls)

        n_location_groups_is_fixed = False

        # determine the number of groups needed
        if dummy_location.n_locations is None and dummy_location.n_agents is None:
            n_location_groups = 1
            groups: list[list] = [[]]

        elif dummy_location.n_locations is None and dummy_location.n_agents is not None:
            if dummy_location.overcrowding is None:
                round_function = round
            elif dummy_location.overcrowding is True:
                round_function = math.floor
            elif dummy_location.overcrowding is False:
                round_function = math.ceil
            else:
                # TODO
                raise Exception

            n_location_groups = max(
                round_function(len(agents) / dummy_location.n_agents),
                1,
            )
            groups: list[list] = [[]]

        elif dummy_location.n_locations is not None and dummy_location.n_agents is None:
            n_location_groups = dummy_location.n_locations
            location_cls.n_agents = max(
                math.floor(len(agents) / n_location_groups),
                1,
            )
            groups: list[list] = [[] for _ in range(n_location_groups)]

        elif dummy_location.n_locations is not None and dummy_location.n_agents is not None:
            n_location_groups = dummy_location.n_locations
            n_location_groups_is_fixed = True
            groups: list[list] = [[]]

        else:
            # TODO:
            raise Exception

        stick_values = {self._get_stick_value(agent, dummy_location) for agent in agents}

        # dummy_location = self._create_dummy_location(location_cls)

        # for each group of sticky agents
        for stick_value in stick_values:
            sticky_agents = [
                agent
                for agent in agents
                if self._get_stick_value(agent, dummy_location) == stick_value
            ]

            assigned = False

            for _, group in enumerate(groups):
                # if there are still enough free places available
                if (dummy_location.n_agents is None) or dummy_location.n_agents - len(group) >= len(
                    sticky_agents
                ):
                    # if sum(
                    #    [dummy_location.find(agent) for agent in sticky_agents],
                    # ) == len(sticky_agents):
                    #    # assign agents
                    for agent in sticky_agents:
                        group.append(agent)
                        dummy_location.add_agent(agent)

                    assigned = True
                    break

            if not assigned:
                if len(groups) < n_location_groups:
                    new_group = []
                    dummy_location = self._create_dummy_location(location_cls)
                    # assign agents
                    for agent in sticky_agents:
                        new_group.append(agent)
                        dummy_location.add_agent(agent)

                    groups.append(new_group)

                else:
                    if not dummy_location.only_exact_n_agents and not n_location_groups_is_fixed:
                        for agent in sticky_agents:
                            groups[overcrowding_i].append(agent)
                        overcrowding_i = (overcrowding_i + 1) % len(groups)

        if dummy_location.only_exact_n_agents:
            groups = [group for group in groups if len(group) == dummy_location.n_agents]

        return groups

    def _get_split_value_affiliated_agents(
        self,
        agents: list,
        split_value: int | str,
    ) -> list:
        group_affiliated_agents = [
            agent for agent in agents if split_value in agent._P2NTEMP_split_values
        ]

        return group_affiliated_agents

    def _get_melted_groups(self, agents: list, location_cls) -> list[list]:
        dummy_location = self._create_dummy_location(location_cls)

        # get all mother locations the agents are nested in
        all_mother_group_ids = {
            self._get_mother_group_id(agent, dummy_location) for agent in agents
        }

        # for each mother location
        for mother_group_id in all_mother_group_ids:
            # get agents that are part of this location
            nested_agents = [
                agent
                for agent in agents
                if self._get_mother_group_id(agent, dummy_location) == mother_group_id
            ]

            # a list that stores a list of groups for each location
            # [
            # [[_agent], [_agent], [_agent]], # groups of location 1
            # [[_agent], [_agent]],           # groups of location 2
            # ]
            groups_to_melt_by_location: list[list[list]] = []

            # for each location that shall be melted
            for melt_location_cls in dummy_location.melt():
                # create dummy location
                melt_dummy_location = self._create_dummy_location(melt_location_cls)
                if melt_location_cls.n_locations is None:
                    melt_location_cls.n_locations = location_cls.n_locations

                # get all agents that should be assigned to this location
                # filter by melt_location
                melt_location_affiliated_agents = self._get_affiliated_agents(
                    agents=nested_agents,
                    dummy_location=melt_dummy_location,
                )
                # filter by main_location
                melt_location_affiliated_agents = self._get_affiliated_agents(
                    agents=melt_location_affiliated_agents,
                    dummy_location=dummy_location,
                )

                # get all values for which seperated groups/locations should be created
                melt_split_values = self._get_split_values(
                    agents=melt_location_affiliated_agents,
                    dummy_location=melt_dummy_location,
                    allow_nesting=False,
                )

                for agent in melt_location_affiliated_agents:
                    agent._P2NTEMP_melt_location_weight = melt_dummy_location.weight(agent)

                # for each split value: get groups and collect them in one list for all values
                location_groups_to_melt: list[list] = []
                for melt_split_value in melt_split_values:
                    melt_split_value_affiliated_agents = self._get_split_value_affiliated_agents(
                        agents=melt_location_affiliated_agents,
                        split_value=melt_split_value,
                    )
                    location_groups_to_melt.extend(
                        self._get_groups(
                            agents=melt_split_value_affiliated_agents,
                            location_cls=melt_location_cls,
                        ),
                    )

                groups_to_melt_by_location.append(location_groups_to_melt)

            # Melt groups
            all_melted_groups: list[list] = []
            z = sorted(
                [len(groups_to_melt) for groups_to_melt in groups_to_melt_by_location],
                reverse=True if dummy_location.recycle else False,
            )[0]

            for i in range(z):
                melted_group = []
                for groups_to_melt in groups_to_melt_by_location:
                    if len(groups_to_melt) > 0:
                        if dummy_location.recycle:
                            melted_group.extend(groups_to_melt[i % len(groups_to_melt)])
                        else:
                            try:
                                melted_group.extend(groups_to_melt[i])
                            except IndexError:
                                pass

                all_melted_groups.append(melted_group)

        return all_melted_groups

    def create_locations(
        self,
        location_classes: list,
        agents: list | p2n.AgentList | None = None,
        clear: bool = False,
    ) -> p2n.LocationList:
        """Creates location instances and connects them with the given agent population.

        Args:
            location_classes (list): A list of location classes.
            agents (list | p2n.AgentList): A list of agents.
            clear (bool): Should the locations already included in the model be removed?

        Returns:
            p2n.LocationList: A list of locations.
        """
        if clear:
            self.model.remove_locations(self.model.locations)

        if agents is None:
            agents = self.model.agents

        # self._dummy_model = p2n.Model()
        self._dummy_model.add_agents(agents)

        for location_cls in location_classes:
            dummy_location = self._create_dummy_location(location_cls)
            str_location_cls = dummy_location.type
            for agent in agents:
                setattr(agent, str_location_cls, None)
                setattr(agent, str_location_cls + "_assigned", False)
                setattr(agent, str_location_cls + "_id", None)
                setattr(agent, str_location_cls + "_position", None)
                setattr(agent, str_location_cls + "_head", None)
                setattr(agent, str_location_cls + "_tail", None)

        locations = []

        # for each location class
        for location_cls in location_classes:
            for agent in agents:
                agent._P2NTEMP_melt_location_weight = None

            # create location dummy in order to use the location's methods
            dummy_location = self._create_dummy_location(location_cls)

            str_location_cls = dummy_location.type

            # If nxgraph is used do some checks
            if dummy_location.nxgraph is not None:
                if dummy_location.n_agents is not None:
                    msg = """You cannot define location.n_agents if location.nxgraph is used. 
                        It will be set to the number of nodes in location.nxgraph automatically."""
                    warnings.warn(msg)
                location_cls.n_agents = len(list(dummy_location.nxgraph.nodes))
                dummy_location.n_agents = len(list(dummy_location.nxgraph.nodes))

                if dummy_location.overcrowding is True:
                    msg = """You cannot define location.overcrowding if location.nxgraph is used. 
                        It will be set to `False` automatically."""
                    warnings.warn(msg)
                location_cls.overcrowding = False
                dummy_location.n_agents = False

            # if dummy_location.n_agents is not None and dummy_location.n_agents < 1:
            #    msg = (
            #        f"""{str_location_cls}.n_agents must be `None` or an integer greater than 0."""
            #    )
            #    raise Exception(msg)

            # bridge
            if not dummy_location.melt():
                bridge_values = {
                    dummy_location.bridge(agent)
                    for agent in self._get_affiliated_agents(
                        agents=agents, dummy_location=dummy_location
                    )
                    if dummy_location.bridge(agent) is not None
                }

                if len(bridge_values) == 0:
                    pass

                elif len(bridge_values) == 1:
                    msg = f"""{str_location_cls}.bridge() returned only one unique value.
                    {str_location_cls}.bridge() must return at least two unique values in order 
                    to create locations that bring together agents with different values on the 
                    same attribute.
                    """
                    warnings.warn(msg)

                elif len(bridge_values) > 1:
                    if dummy_location.n_agents is not None:
                        msg = f"""You cannot use {str_location_cls}.n_agents and 
                        {str_location_cls}.bridge() at the same time. {str_location_cls}.n_agents
                        is ignored."""
                        warnings.warn(msg)

                    melt_list = []

                    # create one MeltLocation for each bridge_value
                    for bridge_value in bridge_values:

                        def filter(self, agent):
                            return dummy_location.bridge(agent) == self.bridge_value

                        dummy_melt_class = type(
                            f"dummy_meltlocation{str(bridge_value)}",
                            (p2n.MeltLocation,),
                            {
                                "filter": filter,
                                "n_agents": 1,
                                "bridge_value": bridge_value,
                            },
                        )

                        melt_list.append(dummy_melt_class)

                    # set the created MeltLocations as return values of melt()
                    def melt(self):
                        return melt_list

                    location_cls.melt = melt
                    dummy_location = self._create_dummy_location(location_cls)

            if not dummy_location.melt():
                # get all agents that could be assigned to locations of this class
                affiliated_agents = self._get_affiliated_agents(
                    agents=agents,
                    dummy_location=dummy_location,
                )

            else:
                affiliated_agents = []

                for melt_location_cls in dummy_location.melt():
                    melt_dummy_location = self._create_dummy_location(melt_location_cls)
                    affiliated_agents.extend(
                        self._get_affiliated_agents(
                            agents=agents,
                            dummy_location=melt_dummy_location,
                        ),
                    )

            # get all values that are used to split the agents into groups
            split_values = self._get_split_values(
                agents=affiliated_agents,
                dummy_location=dummy_location,
                allow_nesting=True,
            )

            if len(split_values) == 0:
                split_values.append("dummy_split_value")

            group_count = 0

            # for each group split value
            for split_value in split_values:
                split_value_locations = []

                # get all agents with that value
                split_value_affiliated_agents = self._get_split_value_affiliated_agents(
                    agents=affiliated_agents,
                    split_value=split_value,
                )

                # if this location does not glue together other locations
                if not dummy_location.melt():
                    group_lists: list[list] = self._get_groups(
                        agents=split_value_affiliated_agents,
                        location_cls=location_cls,
                    )
                else:
                    group_lists = self._get_melted_groups(
                        agents=split_value_affiliated_agents,
                        location_cls=location_cls,
                    )

                # for each group of agents
                for i, group_list in enumerate(group_lists):
                    group_count += 1

                    dummy_location = self._create_dummy_location(location_cls)
                    dummy_location.agents_ = group_list
                    # dummy_location.add_agents(agents)

                    # dummy_location.group_agents = group_list

                    # get all subgroub values
                    subsplit_values = {
                        agent_subsplit_value
                        for agent in group_list
                        for agent_subsplit_value in utils._to_list(
                            dummy_location._subsplit(agent),
                        )
                    }

                    # for each group of agents assigned to a specific sublocation
                    for j, subsplit_value in enumerate(subsplit_values):
                        # get all subsplit affiliated agents
                        subsplit_affiliated_agents = []

                        # for agent in group_affiliated_agents:
                        for agent in group_list:
                            agent_subsplit_value = utils._to_list(
                                dummy_location._subsplit(agent),
                            )
                            if subsplit_value in agent_subsplit_value:
                                subsplit_affiliated_agents.append(agent)

                        # Build the final location
                        location = location_cls(model=self.model)
                        location.setup()
                        location.split_value = split_value
                        location.subsplit_value = subsplit_value
                        location.group_id = i
                        location.subgroup_id = j

                        split_value_locations.append(location)

                        # Assigning process:
                        for agent in subsplit_affiliated_agents:
                            location.add_agent(agent)

                            weight = (
                                agent._P2NTEMP_melt_location_weight
                                if agent._P2NTEMP_melt_location_weight is not None
                                else location.weight(agent)
                            )

                            location.set_weight(
                                agent=agent,
                                weight=weight,
                            )

                            group_info_str = f"gv={location.split_value},gid={location.group_id}"
                            setattr(agent, str_location_cls, group_info_str)
                            setattr(agent, str_location_cls + "_assigned", True)
                            setattr(agent, str_location_cls + "_id", group_count - 1)
                            setattr(agent, str_location_cls + "_position", group_list.index(agent))
                            setattr(
                                agent,
                                str_location_cls + "_head",
                                True if group_list.index(agent) == 0 else False,
                            )
                            setattr(
                                agent,
                                str_location_cls + "_tail",
                                True if group_list.index(agent) == (len(group_list) - 1) else False,
                            )

                        locations.append(location)

                if (
                    dummy_location.n_locations is not None
                    and not dummy_location.only_exact_n_agents
                ):
                    if len(split_value_locations) < dummy_location.n_locations:
                        for _ in range(
                            int(dummy_location.n_locations - len(split_value_locations))
                        ):
                            location = location_cls(model=self.model)
                            location.setup()
                            location.split_value = split_value
                            location.subsplit_value = None
                            location.group_id = None
                            location.subgroup_id = None

                            locations.append(location)

        locations = p2n.LocationList(model=self.model, objs=locations)

        # execute an action after all locations have been created
        for location in locations:
            location.refine()

        # delete temporary agent attributes
        for agent in self._dummy_model.agents:
            for attr in self._temp_agent_attrs:
                if hasattr(agent, attr):
                    delattr(agent, attr)

        return locations

    def create(
        self,
        df: pd.DataFrame,
        location_classes: list,
        agent_class: type[p2n.Agent] = p2n.Agent,
        agent_class_attr: None | str = None,
        agent_class_dict: None | dict = None,
        n_agents: int | None = None,
        sample_level: str | None = None,
        sample_weight: str | None = None,
        replace_sample_level_column: bool = True,
        clear: bool = False,
    ) -> tuple:
        """Creates agents and locations based on a given dataset.

        Combines the Creator-methods `draw_sample()`, `create_agents()` and `create_locations()`.

        Args:
            df (pd.DataFrame): A data set with individual data that forms the basis for
                the creation of agents. Each row is (potentially) translated into one agent.
                Each column is translated into one agent attribute.
            agent_class (type[p2n.Agent]): The class from which the agent instances are created.
            location_classes (list): A list of classes from which the location instances are
                created.
            n_agents (Optional[int], optional): The number of agents that will be created.
                If `n_agents` is set to None, each row of `df` is translated into exactly one agent.
                Otherwise, rows are randomly drawn (with replacement,
                if `n_agents > len(df)`) from `df` until the number of agents created
                equals `n_agents`.
            sample_level (Optional[str], optional): If `sample_level` is None,
                the rows are sampled individually.
                Otherwise the rows are sampled as groups. `sample_level` defines
                which column of `df` contains the group id.
            sample_weight (Optional[str]): The column of df in which should be used as probability
                weight during sampling.
            replace_sample_level_column (bool): Should the original values of the sample level be
                overwritten by unique values after sampling to avoid duplicates?
            clear (bool): Should the agents and locations already included in the model be removed?

        Returns:
            tuple: A list of agents and a list of locations.
        """
        # draw a sample from dataset
        df_sample = self.draw_sample(
            df=df,
            n=n_agents,
            sample_level=sample_level,
            sample_weight=sample_weight,
            replace_sample_level_column=replace_sample_level_column,
        )

        # create agents
        agents = self.create_agents(
            df=df_sample,
            agent_class=agent_class,
            agent_class_attr=agent_class_attr,
            agent_class_dict=agent_class_dict,
            clear=clear,
        )

        # create locations
        locations = self.create_locations(
            agents=agents,
            location_classes=location_classes,
            clear=clear,
        )

        return agents, locations

    def get_df_agents(
        self,
        columns: None | list[str] = None,
        drop_agentpy_columns: bool = True,
    ) -> pd.DataFrame:
        """Returns the latest created population of agents as a dataframe.

        Args:
            columns (list | None): A list of column names that sould be kept.
                All other columns are deleted.
            drop_agentpy_columns (bool): Deletes some columns created by AgentPy.

        Raises:
            Pop2netException: _description_

        Returns:
            pd.DataFrame: A dataframe which contains one row for each
            agent and one column for each agent attribute.
        """
        if self.agents is None:
            msg = "There are no agents."
            raise Pop2netException(msg)

        df = pd.DataFrame([vars(agent) for agent in self.agents])

        if drop_agentpy_columns:
            df = df.drop(
                columns=[
                    "_var_ignore",
                    "id",
                    "type",
                    "log",
                    "model",
                    "p",
                ],
            )

        if columns is not None:
            df = df.loc[:, columns]

        return df



"""Faker of SOEP Data.

SOEP is the socio-economic panel. A long lifed panel study.
"""
# TODO: extend docstring.

from typing import Optional

import faker
import numpy as np
import pandas as pd

from .soep_provider import SOEPProvider


def soep(size: int, seed: Optional[int] = None) -> pd.DataFrame:
    """Create a pandas DataFrame with faked SOEP data.

    Args:
        size (int): Minimum number of people in the dataset.
        seed (Optional[int], optional): Optional seed. Defaults to None.

    Returns:
        pd.DataFrame: A fake SEOP dataset.
    """
    np.random.seed(seed)
    faker.Faker.seed(seed)
    fake = faker.Faker()
    fake.add_provider(SOEPProvider)

    # create correlated variables, dependent on child vs. adult
    cases: list[dict] = []
    try:
        while len(cases) < size:
            for person in fake.household_persons():
                person["pid"] = fake.unique.random_int(min=1, max=size * 10)
                cases.append(person)
                if len(cases) == size:
                    raise StopIteration
    except StopIteration:
        pass

    df = pd.DataFrame(cases)
    return df



"""Base class to create Agent objects."""

from __future__ import annotations

import typing
import warnings

import agentpy as ap

if typing.TYPE_CHECKING:
    from . import location as _location
    from . import sequences as _sequences


class Agent(ap.Agent):
    """This is a Base class to represent agents in the simulation.

    Agents' behavior can be implemented in classes that inherit from this.

    Examples:
        For instance, agents could all be instatiated with the `is_infected` attribute set to
        false::

            class InfectionAgent(Agent):
                def setup(self):
                    self.is_infected = False
    """

    def __init__(self, model, *args, **kwargs) -> None:
        """Agent Constructor.

        All parameters will be passed to the :class:`agentpy.Agent` parent.
        """
        super().__init__(model, *args, **kwargs)

        self.model = model

        self.model.add_agent(self)
        self.setup()

    def setup(self) -> None:
        """Instantiate an Agent.

        This is executed on the instantiation of each agent.
        """

    def neighbors(self, location_classes: list | None = None) -> ap.AgentList:
        """Return all neighbors of an agent.

        Convenience method that returns all neighbors over all locations this agent is currently
        located in. The locations to be considered can be defined with location_classes.

        Args:
            location_classes: A list of location_classes.

        Returns:
            All agents co-located with this agent over all locations.
        """
        return self.model.neighbors_of_agent(self, location_classes=location_classes)

    def shared_locations(self, agent, location_classes: list | None = None):
        return self.model.locations_between_agents(
            agent1=self,
            agent2=agent,
            location_classes=location_classes,
        )

    def add_location(self, location: _location.Location) -> None:
        """Add this Agent to a given location.

        Args:
            location: Add agent to this location.
        """
        self.model.add_agent_to_location(self, location)

    def add_locations(self, locations: list) -> None:
        for location in locations:
            self.add_location(location)

    def remove_location(self, location: _location.Location) -> None:
        """Remove this Agent from a given location.

        Args:
            location: Remove agent from this location.
        """
        self.model.remove_agent_from_location(self, location)

    def remove_locations(self, locations: list) -> None:
        for location in locations:
            self.remove_location(location)

    @property
    def locations(self) -> _sequences.LocationList:
        """Return a list of locations that this agent is associated with.

        Returns:
            A list of locations.
        """
        return self.model.locations_of_agent(self)

    def get_agent_weight(self, agent: Agent, location_classes: list | None = None) -> float:
        """Return the contact weight between this agent and a given other agent.

        This is summed over all shared locations.

        Args:
            agent_v: The other agent.

        Returns:
            A weight of the contact between the two agents.
        """
        weight = 0
        for location in self.shared_locations(agent=agent, location_classes=location_classes):
            weight += location.project_weights(agent1=self, agent2=agent)
        return weight

    def get_location_weight(self, location) -> float:
        return self.model.get_weight(agent=self, location=location)

    def connect(self, agent: Agent, location_cls: type):
        """Connects this agent with a given other agent via an instance of a given location class.

        Args:
            agents (list): An agent to connect with.
            location_cls (type): The location class that is used to create a location instance.
        """
        self.model.connect_agents(agents=[self, agent], location_cls=location_cls)

    def disconnect(
        self,
        neighbor: Agent,
        location_classes: list | None = None,
        remove_self=True,
        remove_neighbor=True,
        remove_locations: bool = False,
    ):
        """Disconnects this agent from a given other agent by removing them from shared locations.

        If a list of location types is given, only shared locations of the given types are
        considered. Turn on `remove_locations` in order to not only remove the agents from the
        given location instance but also to remove the location instance from the model.  Keep in
        mind that this may affect other agents that are still connected with the location instance.

        Args:
            neighbor (Agent): An agent to disconnect from.
            location_classes (list | None, optional): A list of location types to specify which
            shared locations are considered. Defaults to None.
            remove_self (bool): Should the agent be removed from the shared locations?
                Defaults to True.
            remove_neighbor (bool): Should the neighbor be removed from the shared locations?
                Defaults to True.
            remove_locations (bool, optional): A bool that determines whether the shared locations
                shall be removed from the model. Defaults to False.
        """
        shared_locations = self.shared_locations(
            agent=neighbor,
            location_classes=location_classes,
        )

        for location in shared_locations:
            warn = False
            for agent in location.agents:
                if agent not in [self, neighbor]:
                    warn = True
                    break

            if remove_self:
                location.remove_agent(self)

                if warn:
                    msg = (
                        "There are other agents at the location from which you have removed agents."
                    )
                    warnings.warn(msg)

            if remove_neighbor:
                location.remove_agent(neighbor)

                if warn:
                    msg = (
                        "There are other agents at the location from which you have removed agents."
                    )
                    warnings.warn(msg)

            if remove_locations:
                self.model.remove_location(location)

                if warn:
                    msg = "You have removed a location to which other agents were still connected."
                    warnings.warn(msg)



"""A module that helps with creating insights into the network."""

from __future__ import annotations

import typing

from bokehgraph import BokehBipartiteGraph
from bokehgraph import BokehGraph
import networkx as nx
import pandas as pd
import seaborn as sns
from tabulate import tabulate

import pop2net.utils as utils

if typing.TYPE_CHECKING:
    from . import location as _location


class NetworkInspector:
    """Helper class that contains functions to inspect the network of a pop2net model."""

    def __init__(self, model) -> None:
        """Initiate a NetworkInspector.

        Args:
            model (Model): The model.
        """
        self.model = model

    def plot_bipartite_network(
        self,
        agent_attrs: list | None = None,
        location_attrs: list | None = None,
        agent_color: str | None = None,
        location_color: str | None = None,
        edge_alpha: str = "weight",
        edge_color: str = "black",
    ) -> None:
        """Plots the two-mode network of agents and locations.

        Args:
            agent_color (str, optional): The agent attribute that determines the
                color of the agent nodes.
            agent_attrs (list | None, optional): A list of agent attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            location_color (str, optional): The location attribute that determines the
                color of the location nodes.
            location_attrs (list | None, optional): A list of location attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            edge_alpha (str, optional): The node attribute that determines the edges' transparency.
                Defaults to "weight".
            edge_color (str, optional): The node attribute that determines the edges' color.
                Defaults to "black".
            include_0_weights (bool, optional): Should edges with a weight of zero be included in
                the plot? Defaults to True.
        """
        if agent_attrs is None:
            agent_attrs = ["type"]
        else:
            agent_attrs = list(agent_attrs)
            if "type" not in agent_attrs:
                agent_attrs.append("type")

        if agent_color is not None and agent_color not in agent_attrs:
            agent_attrs.append(agent_color)

        if location_attrs is None:
            location_attrs = ["type"]
        else:
            location_attrs = list(location_attrs)
            if "type" not in location_attrs:
                location_attrs.append("type")

        if location_color is not None and location_color not in location_attrs:
            location_attrs.append(location_color)

        graph = self.model.export_bipartite_network(
            agent_attrs=agent_attrs,
            location_attrs=location_attrs,
        )

        graph_layout = nx.drawing.spring_layout(graph)
        plot = BokehBipartiteGraph(graph, width=400, height=400, hover_edges=True)
        plot.layout(layout=graph_layout)
        plot.draw(
            node_color_lv0="firebrick" if agent_color is None else agent_color,
            node_color_lv1="black" if location_color is None else location_color,
            edge_alpha=edge_alpha,
            edge_color=edge_color,
            # node_palette="random",
        )

    def plot_agent_network(
        self,
        agent_color: str | None = None,
        agent_attrs: list | None = None,
        edge_alpha: str = "weight",
        edge_color: str = "black",
        include_0_weights: bool = True,
    ) -> None:
        """Plots the agent network.

        Args:
            agent_color (str, optional): The node attribute that determines the
                color of the nodes. If None, the node color represents whether
                it is a location or an agent instance.
            agent_attrs (list | None, optional): A list of agent attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            edge_alpha (str, optional): The node attribute that determines the edges' transparency.
                Defaults to "weight".
            edge_color (str, optional): The node attribute that determines the edges' color.
                Defaults to "black".
            include_0_weights (bool, optional): Should edges with a weight of zero be included in
                the plot? Defaults to True.
        """
        if agent_attrs is None:
            agent_attrs = ["type"]
        else:
            agent_attrs = list(agent_attrs)
            if "type" not in agent_attrs:
                agent_attrs.append("type")

        if agent_color is not None and agent_color not in agent_attrs:
            agent_attrs.append(agent_color)

        graph = self.model.export_agent_network(
            node_attrs=agent_attrs,
            include_0_weights=include_0_weights,
        )

        graph_layout = nx.drawing.spring_layout(graph)
        plot = BokehGraph(graph, width=400, height=400, hover_edges=True)
        plot.layout(layout=graph_layout)
        plot.draw(
            node_color="firebrick" if agent_color is None else agent_color,
            edge_alpha=edge_alpha,
            edge_color=edge_color,
        )

    def plot_networks(
        self,
        agent_attrs: list | None = None,
        location_attrs: list | None = None,
        agent_color: str | None = None,
        location_color: str | None = None,
        edge_alpha: str = "weight",
        edge_color: str = "black",
        include_0_weights: bool = True,
    ) -> None:
        """Plots the two-mode network and the agent network.

        Args:
            agent_color (str, optional): The agent attribute that determines the
                color of the agent nodes.
            agent_attrs (list | None, optional): A list of agent attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            location_color (str, optional): The location attribute that determines the
                color of the location nodes.
            location_attrs (list | None, optional): A list of location attributes that
                should be shown as node attributes in the network graph. Defaults to None.
            edge_alpha (str, optional): The node attribute that determines the edges' transparency.
                Defaults to "weight".
            edge_color (str, optional): The node attribute that determines the edges' color.
                Defaults to "black".
            include_0_weights (bool): Should edges with a weight of zero be included in
                the plot? Defaults to True.
        """
        self.plot_bipartite_network(
            agent_color=agent_color,
            agent_attrs=agent_attrs,
            location_color=location_color,
            location_attrs=location_attrs,
            edge_color=edge_color,
            edge_alpha=edge_alpha,
            # include_0_weights=include_0_weights,
        )

        self.plot_agent_network(
            agent_color=agent_color,
            agent_attrs=agent_attrs,
            edge_color=edge_color,
            edge_alpha=edge_alpha,
            include_0_weights=include_0_weights,
        )

    def eval_affiliations(self, return_data=False) -> tuple[pd.DataFrame, pd.DataFrame] | None:
        """Prints information on the distribution of agents per location and locations per agent.

        Raises:
            Pop2netException: _description_
            Pop2netException: _description_
        """
        df1 = pd.DataFrame(
            [
                {
                    "location_class": str(type(location)).split(".")[-1].split("'")[0],
                    "n_agents": len(location.agents),
                }
                for location in self.model.locations
            ],
        )

        df1 = df1.groupby("location_class").describe()
        df1.columns = df1.columns.droplevel()
        df1 = df1.drop("count", axis=1)

        utils.print_header("Number of agents per location")
        print(df1)

        df2 = pd.DataFrame(
            [
                {
                    "agent_id": agent.id,
                    "n_affiliated_locations": len(agent.locations),
                }
                for agent in self.model.agents
            ],
        )
        df2 = df2.n_affiliated_locations.describe()
        df2 = df2.drop("count", axis=0)

        utils.print_header("Number of affiliated locations per agent")
        print(df2)

        if return_data:
            return df1, df2
        return None

    # TODO: calculate relative freqs
    def create_contact_matrix(
        self,
        agents: list | None = None,
        attr: str = "id",
        weighted: bool = False,
        plot: bool = True,
        annot: bool = False,
        return_df: bool = False,
    ) -> pd.DataFrame:
        """Create a contact matrix as a DataFrame from a given model's agent list.

        Args:
            agents: A list of agents.
            attr: The agent attribute which is shown in the matrix.
            weighted: Should the contacts be weighted? Defaults to False.
            plot: Should the matrix be plotted? Defaults to False.
            annot: Should the plottet matrix be annotated? Defaults to False.
            return_df: Should the data be returned as pandas.DataFrame?

        Returns:
            A DataFrame containing a contact matrix based on `attr`.
        """
        if agents is None:
            agents = self.model.agents

        contact_data = []
        attr_values = []
        pairs = []

        attr_u_name = f"{attr}_u"
        attr_v_name = f"{attr}_v"

        for agent_u in agents:
            attr_u = getattr(agent_u, attr)
            if attr_u is not None:
                attr_values.append(attr_u)

                for agent_v in agent_u.neighbors():
                    attr_v = getattr(agent_v, attr)
                    if attr_v is not None:
                        attr_values.append(attr_v)

                        pair = {agent_u.id, agent_v.id}

                        if pair not in pairs:
                            contact_data.append(
                                {
                                    "id_u": agent_u.id,
                                    attr_u_name: attr_u,
                                    "id_v": agent_v.id,
                                    attr_v_name: attr_v,
                                    "weight": agent_u.get_agent_weight(agent_v),
                                },
                            )
                            pairs.append(pair)

        attr_values = list(set(attr_values))

        df = pd.DataFrame(index=sorted(attr_values, reverse=True), columns=sorted(attr_values))
        df = df.fillna(0)

        weight_total = 0
        for contact in contact_data:
            weight = contact["weight"] if weighted else 1
            weight_total += weight

            df.loc[contact[attr_u_name], contact[attr_v_name]] = (
                df.loc[contact[attr_u_name], contact[attr_v_name]] + weight
            )
            df.loc[contact[attr_v_name], contact[attr_u_name]] = (
                df.loc[contact[attr_v_name], contact[attr_u_name]] + weight
            )

        if plot:
            g = sns.heatmap(df, annot=annot, vmin=0, fmt="g")
            g.set(xlabel=attr, ylabel=attr)

        if return_df:
            return df

    def network_measures(self, node_attrs=None) -> dict | list[dict]:
        """Creates nx networkgraph and calculates common network measures.

        If the created network consist of independent groups of nodes
        subgraphs are created and measures are calculated for each subgraph

        Args:
            node_attrs: A list of agent attributes

        Return:
            dictionary/or list of dictionaries of the common network
            measure results
        """
        result_dict = {}
        nx_graph = self.model.export_agent_network(node_attrs=node_attrs)

        # make distinction between multiple independent networks and one network
        if nx.is_connected(nx_graph):
            result_dict["diameter"] = nx.diameter(nx_graph, weight="weight")
            result_dict["density"] = nx.density(nx_graph)
            result_dict["transitivity"] = nx.transitivity(nx_graph)
            result_dict["avg_clustering"] = nx.average_clustering(
                nx_graph,
                weight="weight",
            )
            result_dict["avg_path_length"] = nx.average_shortest_path_length(
                nx_graph,
                weight="weight",
            )
            # result_dict["periphery"] = nx.periphery(nx_graph, weight = "weight")
            # result_dict["center"] = nx.center(nx_graph, weight = "weight")
            # result_dict["centrality"] = nx.degree_centrality(nx_graph)
            return result_dict
        else:
            # sort subgraph component size(=num of nodes) in ascending order
            component_list = sorted(
                nx.connected_components(nx_graph),
                key=len,
                reverse=False,
            )

            # create graph for each component and calculate network measures
            result_list = []
            for component in component_list:
                result_dict_subgraph = {}
                try:
                    nx_subgraph = nx_graph.subgraph(component)
                except nx.NetworkXError:
                    print("Cant make graph out of component")
                    break

                result_dict_subgraph["diameter"] = nx.diameter(
                    nx_subgraph,
                    weight="weight",
                )
                result_dict_subgraph["density"] = nx.density(nx_subgraph)
                result_dict_subgraph["transitivity"] = nx.transitivity(nx_subgraph)
                result_dict_subgraph["avg_clustering"] = nx.average_clustering(
                    nx_subgraph,
                    weight="weight",
                )
                result_dict_subgraph["avg_path_length"] = nx.average_shortest_path_length(
                    nx_subgraph,
                    weight="weight",
                )

                # result_dict_subgraph["centrality"] = nx.degree_centrality(nx_subgraph)
                # result_dict_subgraph["periphery"] = nx.periphery(nx_subgraph, weight = "weight")
                # result_dict_subgraph["center"] = nx.center(nx_subgraph, weight = "weight")
                result_list.append(result_dict_subgraph)
            return result_list

    ## TODO wie gebe ich hier den richtigen Datentyp an???
    def location_crosstab(
        self,
        select_locations: _location.location | list[_location.Location],
        agent_attributes: str | list[str],
        output_format="table",
    ) -> list[pd.DataFrame] | None:
        """Crosstable for specified location classes and agent attribute."""
        # Make every Parameter a list
        if select_locations:
            if not isinstance(select_locations, list):
                select_locations = [select_locations]

        if agent_attributes:
            if not isinstance(agent_attributes, list):
                agent_attributes = [agent_attributes]

        # determine eligible locations classes
        valid_locations = []
        if select_locations:
            for location_instance in self.model.locations:
                for locationtype in select_locations:
                    if isinstance(location_instance, locationtype):
                        valid_locations.append(location_instance)
        else:
            valid_locations = list(self.model.locations)

        # create agent df per location instance
        if output_format == "table":
            agent_dfs = {}
            if agent_attributes:
                for i, location_instance in enumerate(valid_locations):
                    title = f'{i+1}.Location: {str(location_instance).split(" ")[0]}'
                    df = pd.DataFrame([vars(agent) for agent in location_instance.agents])
                    df = df[list(agent_attributes)]
                    agent_dfs[title] = df

            for title, df in agent_dfs.items():
                for agent_attribute in agent_attributes:
                    crosstab_table = pd.crosstab(
                        index=df[agent_attribute],
                        columns="count",
                    )

                    print(f"{title}")
                    print(
                        tabulate(
                            crosstab_table,
                            headers=[agent_attribute, "count"],
                            tablefmt="fancy_grid",
                        ),
                    )
                    print("\n")
        if output_format == "df":
            agent_dfs = {}
            result_list = []
            for i, location_instance in enumerate(valid_locations):
                title = f'{i+1}.Location: {str(location_instance).split(" ")[0]}'
                location_type = str(location_instance).split(" ")[0]
                df = pd.DataFrame([vars(agent) for agent in location_instance.agents])

                # only keep wanted columns (agent attributes)
                df = df[list(agent_attributes)]

                # add location type as str
                df["location_type"] = location_type

                agent_dfs[title] = df

            for agent_attribute in agent_attributes:
                location_id = 0
                df_list_of_attribute = []
                for _, df in agent_dfs.items():
                    crosstab_table = pd.crosstab(
                        index=df[agent_attribute],
                        columns="count",
                    )
                    crosstab_table.insert(
                        0,
                        "location_id",
                        [location_id] * len(crosstab_table.index),
                    )
                    crosstab_table["location_type"] = df["location_type"][0]

                    df_list_of_attribute.append(crosstab_table)

                    location_id += 1
                df_attribute = pd.concat(df_list_of_attribute, ignore_index=False)

                # fix columns after concat
                df_attribute.reset_index(inplace=True)
                df_attribute.columns.names = ["index"]
                location_array = df_attribute["location_id"].copy()
                df_attribute.drop(columns="location_id", inplace=True)
                df_attribute.insert(0, "location_id", location_array)

                result_list.append(df_attribute)
            return result_list
        return None

    def location_information(
        self,
        select_locations: _location.Location | list[_location.Location] | None = None,
        agent_attributes: str | None | list[str] | None = None,
        output_format: str = "table",
    ) -> None | pd.DataFrame:
        """Provides information on the agents assigned to location instances.

           Displayed  information can be filtered by specifying certain location
           classes and agent_attributes

        Args:
            select_locations (p2n.Location | list[p2n.Location] | None, optional): A list of
                location classes. Defaults to None.
            agent_attributes (str | None | list[str] | None, optional): A list of agent attributes.
                Defaults to None.
            output_format (str, optional): A str determining what is returned. Defaults to "table".

        Returns:
            None | pd.DataFrame: A pandas.DataFrame or nothing.
        """
        if select_locations:
            if not isinstance(select_locations, list):
                select_locations = [select_locations]

        if agent_attributes:
            if not isinstance(agent_attributes, list):
                agent_attributes = [agent_attributes]

        # determine eligible locations classes
        valid_locations = []
        if select_locations:
            for location_instance in self.model.locations:
                for locationtype in select_locations:
                    if isinstance(location_instance, locationtype):
                        valid_locations.append(location_instance)
        else:
            valid_locations = list(self.model.locations)

        # create agent df per location instance
        agent_dfs = {}
        if agent_attributes:
            for i, location_instance in enumerate(valid_locations):
                # Create the title of printout
                title = f'{i+1}.Location: {str(location_instance).split(" ")[0]}'
                location_type = str(location_instance).split(" ")[0]
                # get all agents per location instance, subset df by agent-attributes
                df = pd.DataFrame([vars(agent) for agent in location_instance.agents])
                df = df[list(agent_attributes)]
                df["location_type"] = location_type
                agent_dfs[title] = df
        else:
            for i, location_instance in enumerate(valid_locations):
                title = f'{i+1}.Location: {str(location_instance).split(" ")[0]}'
                location_type = str(location_instance).split(" ")[0]
                df = pd.DataFrame([vars(agent) for agent in location_instance.agents])
                df.drop(df.iloc[:, 0:7], axis=1, inplace=True)
                df["location_type"] = location_type
                agent_dfs[title] = df

        if output_format == "table":
            #### Print Part "Basic"
            for title, df in agent_dfs.items():
                print(f"{title} \n")
                print(tabulate(df, headers="keys", tablefmt="fancy_grid"))
                print("\n")

        if output_format == "df":
            location_id_counter = 0
            df_list = []
            for _, df in agent_dfs.items():
                df.insert(0, "location_id", [location_id_counter] * len(df.index))
                location_id_counter += 1
                df_list.append(df)
            return pd.concat(df_list, ignore_index=True)
        return None



"""Pop2net. An extension package for location-based simulations based on AgentPy."""

from agentpy import AgentList

from .agent import Agent
from .creator import Creator
from .exceptions import Pop2netException
from .inspector import NetworkInspector
from .location import Location
from .location import MagicLocation
from .location import MeltLocation
from .model import Model
from .sequences import LocationList

__all__ = [
    "AgentList",
    "Agent",
    "Pop2netException",
    "Location",
    "MagicLocation",
    "MeltLocation",
    "NetworkInspector",
    "Model",
    "LocationList",
    "Creator",
]



"""The model class. It encapsulates the full simulation."""

from __future__ import annotations

import itertools
import typing
import warnings

import agentpy as ap
from agentpy import AgentList
import networkx as nx

if typing.TYPE_CHECKING:
    from . import agent as _agent
    from . import location as _location

from pop2net.sequences import LocationList
import pop2net.utils as utils


class Model(ap.Model):
    """Class the encapsulates a full simluation.

    This very closely follows the logic of the :class:`agentpy.Model` package. See
    :class:`agentpy.Model` for more information.
    """

    def __init__(self, parameters=None, _run_id=None, **kwargs):
        """Initiate a simulation.

        Args:
            parameters (dict, optional): An optional parameter dict that is passed to
            :class:`agentpy.Model`. Defaults to None.
            _run_id (int, optional): An optional _run_id that is passed to :class:`agentpy.Model`.
            Defaults to None.
            **kwargs: Optional parameters that are all passed to :class:`agentpy.Model`.
        """
        super().__init__(parameters, _run_id, **kwargs)
        self.g = nx.Graph()

    def sim_step(self) -> None:
        """Do 1 step in the simulation."""
        self.t += 1

        # TODO: Rethink the following:
        for location in self.locations:
            if hasattr(location, "static_weight") and hasattr(location, "_update_weights"):
                if not location.static_weight:
                    location._update_weights()

        self.step()
        self.update()

        if self.t >= self._steps:  # type: ignore
            self.running = False

    @property
    def agents(self) -> AgentList:
        """Show a iterable view of all agents in the environment.

        Returns:
            AgentList: A non-mutable AgentList of all agents in the environment.
        """
        return AgentList(
            model=self.model,
            objs=[data["_obj"] for _, data in self.g.nodes(data=True) if data["bipartite"] == 0],
        )

    @property
    def agents_by_id(self) -> dict:
        """Returns a dictionary which stores the model's agents by their id.

        Returns:
            dict: A dictionary which stores the model's agents by their id.
        """
        return {agent.id: agent for agent in self.agents}

    @property
    def locations_by_id(self) -> dict:
        """Returns a dictionary which stores the model's locations by their id.

        Returns:
            dict: A dictionary which stores the model's locations by their id.
        """
        return {location.id: location for location in self.locations}

    @property
    def locations(self) -> LocationList:
        """Show a iterable view of all locations in the environment.

        Returns:
            LocationList: a non-mutable LocationList of all locations in the environment.
        """
        return LocationList(
            model=self.model,
            objs=[data["_obj"] for _, data in self.g.nodes(data=True) if data["bipartite"] == 1],
        )

    def add_agent(self, agent: _agent.Agent) -> None:
        """Add an agent to the environment.

        The added agent will have no connections to other agents or locatons by default.
        If the agent is already in the current environment, this methods does nothing.

        Args:
            agent: Agent to be added to the environment.
        """
        if not self.g.has_node(agent.id):
            self.g.add_node(agent.id, bipartite=0, _obj=agent)

    def add_agents(self, agents: list) -> None:
        """Add agents to the environment.

        Args:
            agents (list): A list of the agents to be added.
        """
        for agent in agents:
            self.add_agent(agent)

    def add_location(self, location: _location.Location) -> None:
        """Add a location to the environment.

        The added location will have no connections to other agents or locatons by default.
        If the location is already in the current environment, this methods does nothing.

        Args:
            location: Location to be added to the environment.
        """
        if not self.g.has_node(location.id):
            self.g.add_node(location.id, bipartite=1, _obj=location)

    def add_locations(self, locations: list) -> None:
        """Add multiple locations to the environment at once.

        Args:
            locations (list): An iterable over multiple locations.
        """
        for location in locations:
            self.add_location(location)

    def add_agent_to_location(
        self,
        location: _location.Location,
        agent: _agent.Agent,
        weight: float = 1,
        **kwargs,
    ) -> None:
        """Add an agent to a specific location.

        Both the agent and the location have to be defined beforehand. All additional keyword
        arguments will be edge attributes for this connection.

        Args:
            location: Location the agent is to be added to.
            agent: Agent to be added to the location.
            weight: An optional weight for the connection.
            **kwargs: Additional edge attributes.

        Raises:
            Exception: Raised if the location does not exist in the environment.
            Exception: Raised if the agent does not exist in the environment.
        """
        # TODO: Create custom exceptions
        if not self.g.has_node(location.id):
            msg = f"Location {location} does not exist in Environment!"
            raise Exception(msg)
        if not self.g.has_node(agent.id):
            msg = f"Agent {agent} does not exist in Environment!"
            raise Exception(msg)

        self.g.add_edge(agent.id, location.id, **kwargs)
        self.set_weight(agent=agent, location=location, weight=weight)

    def remove_agent(self, agent: _agent.Agent) -> None:
        """Remove an agent from the environment.

        If the agent does not exist in the environment, this method does nothing.

        Args:
            agent: Agent to be removed.
        """
        if self.g.has_node(agent.id):
            self.g.remove_node(agent.id)

    def remove_agents(self, agents: list) -> None:
        """Remove multiple agents from the environment at once.

        Args:
            agents (list): An iterable over multiple agents.
        """
        for agent in agents:
            self.remove_agent(agent)

    def remove_location(self, location: _location.Location) -> None:
        """Remove a location from the environment.

        If the location does not exist in the environment, this method does nothing.

        Args:
            location: Location to be removed.
        """
        if self.g.has_node(location.id):
            self.g.remove_node(location.id)

    def remove_locations(self, locations: list) -> None:
        """Remove multiple locations at once.

        Args:
            locations (list): An iterable over locations.
        """
        for location in locations:
            self.remove_location(location)

    def remove_agent_from_location(
        self,
        location: _location.Location,
        agent: _agent.Agent,
    ) -> None:
        """Remove an agent from a location.

        Args:
            location: Location, the agent is to be removed from.
            agent: Agent to be disassociated with the location.

        Raises:
            Exception: Raised if the location does not exist in the environment.
            Exception: Raised if the agent does not exist in the environment.
        """
        # TODO: use custom exceptions
        if not self.g.has_node(location.id):
            msg = f"Location {location} does not exist in Environment!"
            raise Exception(msg)
        if not self.g.has_node(agent.id):
            msg = f"Agent {agent} does not exist in Environment!"
            raise Exception(msg)

        if self.g.has_edge(agent.id, location.id):
            self.g.remove_edge(agent.id, location.id)

    def agents_of_location(self, location: _location.Location) -> AgentList:
        """Return the list of agents associated with a specific location.

        Args:
            location: The desired location.

        Returns:
            A list of agents.
        """
        nodes = self.g.neighbors(location.id)
        return AgentList(
            self.model,
            (self.g.nodes[node]["_obj"] for node in nodes if self.g.nodes[node]["bipartite"] == 0),
        )

    def locations_of_agent(self, agent: _agent.Agent) -> LocationList:
        """Return the list of locations associated with a specific agent.

        Args:
            agent: The desired agent.

        Returns:
            A list of locations.
        """
        nodes = self.g.neighbors(agent.id)
        return LocationList(
            self.model,
            (self.g.nodes[node]["_obj"] for node in nodes if self.g.nodes[node]["bipartite"] == 1),
        )

    def neighbors_of_agent(
        self,
        agent: _agent.Agent,
        location_classes: list | None = None,
    ) -> AgentList:
        """Return a list of neighboring agents for a specific agent.

        The locations to be considered can be defined with location_classes.

        Args:
            agent: Agent of whom the neighbors are to be returned.
            location_classes: A list of location_classes.

        Returns:
            The list of neighbors for the specified agent.
        """
        if location_classes:
            location_classes = [
                (utils._get_cls_as_str(cls) if not isinstance(cls, str) else cls)
                for cls in location_classes
            ]

            locations = (
                node
                for node in self.g.neighbors(agent.id)
                if self.g.nodes[node]["bipartite"] == 1
                and self.g.nodes[node]["_obj"].type in location_classes
            )
        else:
            locations = (
                node for node in self.g.neighbors(agent.id) if self.g.nodes[node]["bipartite"] == 1
            )

        neighbor_agents = {
            agent_id
            for location_id in locations
            for agent_id in self.g.neighbors(location_id)
            if self.g.nodes[agent_id]["bipartite"] == 0
        }
        return AgentList(
            self.model,
            (
                self.g.nodes[agent_id]["_obj"]
                for agent_id in neighbor_agents
                if agent_id != agent.id
            ),
        )

    # TODO: evlt. filtern nach Klasse oder Key einbauen
    def _objects_between_objects(self, object1, object2, object_classes: list | None = None):
        paths = list(
            nx.all_simple_paths(
                G=self.g,
                source=object1.id,
                target=object2.id,
                cutoff=2,
            ),
        )

        objects_between = [self.g.nodes[path[1]]["_obj"] for path in paths]

        if object_classes is not None:
            if len(object_classes) < 1:
                # TODO
                raise Exception

            object_classes = [
                (utils._get_cls_as_str(cls) if not isinstance(cls, str) else cls)
                for cls in object_classes
            ]
            filtered_objects_between = [o for o in objects_between if o.type in object_classes]
            return filtered_objects_between
        else:
            return objects_between

    def locations_between_agents(self, agent1, agent2, location_classes: list | None = None):
        """Return all locations the connect two agents.

        Args:
            agent1 (Agent): Agent 1.
            agent2 (Agent): Agent 2.
            location_classes (tuple, optional): Constrain the locations to the following types.
                Defaults to ().

        Returns:
            LocationList: A list of locations.
        """
        return LocationList(
            model=self.model,
            objs=self._objects_between_objects(agent1, agent2, location_classes),
        )

    def agents_between_locations(self, location1, location2, agent_classes: list | None = None):
        """Return all agents between two locations.

        Args:
            location1 (Location): Location 1.
            location2 (Location): Location 2.
            agent_classes (tuple, optional): Constrain the agents to the following types.
                Defaults to ().

        Returns:
            AgentList: A list of agents.
        """
        return AgentList(
            model=self.model,
            objs=self._objects_between_objects(location1, location2, agent_classes),
        )

    def set_weight(self, agent, location, weight) -> None:
        """Set the weight of an agent at a location.

        Args:
            agent (Agent): The agent.
            location (Location): The location.
            weight (int): The weight
        """
        self.g[agent.id][location.id]["weight"] = 1 if weight is None else weight

    def get_weight(self, agent, location) -> int:
        """Get the weight of an agent at a location.

        Args:
            agent (Agent): The agent.
            location (Location): The location.

        Returns:
            int: The weight.
        """
        return self.g[agent.id][location.id]["weight"]

    def connect_agents(self, agents: list, location_cls: type):
        """Connects multiple agents via an instance of a given location class.

        Args:
            agents (list): A list of agents.
            location_cls (type): The location class that is used to create a location instance.
        """
        location = location_cls(model=self)
        location.add_agents(agents)

    def disconnect_agents(
        self,
        agents: list,
        location_classes: list | None = None,
        remove_locations: bool = False,
    ):
        """Disconnects agents by removing them from shared locations.

        If a list of location types is given, only shared locations of the given types are
        considered. Turn on `remove_locations` in order to not only remove the given agents from the
        given location instance but also to remove the location instance from the model.
        Use this method with care because removing agents from locations also disconnects those
        agents from all other agents connected to the location. Removing the location instance from
        the model could have even more sideeffects to those agents still connected with this
        location!

        Args:
            agents (list): A list of agents.
            location_classes (list | None, optional): A list of location types to specify which
            shared locations are considered. Defaults to None.
            remove_locations (bool, optional): A bool that determines whether the shared locations
                shall be removed from the model. Defaults to False.
        """
        pairs = list(itertools.combinations(agents, 2))

        shared_locations = []

        for agent1, agent2 in pairs:
            shared_locations.extend(
                self.locations_between_agents(
                    agent1=agent1,
                    agent2=agent2,
                    location_classes=location_classes,
                )
            )

        shared_locations = set(shared_locations)

        for location in shared_locations:
            warn = False
            for agent in location.agents:
                if agent not in agents:
                    warn = True
                    break

            if warn:
                msg = "There are other agents at the location from which you have removed agents."
                warnings.warn(msg)

            location.remove_agents(agents)

            if remove_locations:
                self.remove_location(location=location)

                if warn:
                    msg = "You have removed a location to which other agents were still connected."
                    warnings.warn(msg)

    def export_bipartite_network(
        self,
        agent_attrs: list | None = None,
        location_attrs: list | None = None,
    ):
        graph = self.g.copy()

        for i in graph:
            if graph.nodes[i]["bipartite"] == 0:
                if agent_attrs is not None:
                    for agent_attr in agent_attrs:
                        graph.nodes[i][agent_attr] = getattr(graph.nodes[i]["_obj"], agent_attr)

            elif graph.nodes[i]["bipartite"] == 1:
                if location_attrs is not None:
                    for location_attr in location_attrs:
                        graph.nodes[i][location_attr] = getattr(
                            graph.nodes[i]["_obj"], location_attr
                        )

            del graph.nodes[i]["_obj"]
        return graph

    def export_agent_network(
        self,
        node_attrs: list | None = None,
        include_0_weights: bool = True,
    ) -> nx.Graph:
        """Creates a projection of the model's bipartite network.

        Args:
            node_attrs: A list of agent attributes
            include_0_weights: Should edges with weight 0 be displayed?

        Returns:
            A weighted graph created from a model's agent list. Agents are connected if they are
            neighbors in the model. Their connecting edge include the contact_weight as "weight"
            attribute.
        """
        graph = nx.Graph()

        # create nodes
        for agent in self.agents:
            if not graph.has_node(agent.id):
                node_attr_dict = (
                    {node_attr: vars(agent)[node_attr] for node_attr in node_attrs}
                    if node_attrs is not None
                    else {}
                )
                graph.add_node(agent.id, **node_attr_dict)

        # create edges
        for agent in self.agents:
            for agent_v in agent.neighbors():
                if not graph.has_edge(agent.id, agent_v.id):
                    weight = agent.get_agent_weight(agent_v)
                    if include_0_weights or weight > 0:
                        graph.add_edge(agent.id, agent_v.id, weight=weight)

        return graph



# ruff: noqa: D102, D105, D107
"""Agentpy Lists Module.

Content: Lists for objects, environments, and agents.

BSD 3-Clause License

Copyright (c) 2020-2021 Joël Foramitti

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import annotations

from collections.abc import Sequence

from . import location


class LocationSequence:
    """Base class for agenpty sequences."""

    def __repr__(self):
        """Return a REPL representation of LocationSequence.

        Returns:
            str: REPL representation.
        """
        len_ = len(list(self))
        s = "s" if len_ != 1 else ""
        return f"{type(self).__name__} ({len_} object{s})"

    def __getattr__(self, name):
        """Return callable list of attributes."""
        if name[0] == "_":  # Private variables are looked up normally
            # Gives numpy conversion correct error for __array_struct__ lookup
            super().__getattr__(name)
        else:
            return AttrIter(self, attr=name)

    def _set(self, key, value):
        object.__setattr__(self, key, value)

    @staticmethod
    def _obj_gen(model, n, cls, *args, **kwargs):
        """Generate objects for sequence."""
        if cls is None:
            cls = location.Location

        if args != ():
            msg = (
                "Sequences no longer accept extra arguments without a keyword. "
                f"Please assign a keyword to the following arguments: {args}"
            )
            raise Exception(
                msg,
            )

        for i in range(n):
            # AttrIter values get broadcasted among agents
            i_kwargs = {
                k: arg[i] if isinstance(arg, AttrIter) else arg for k, arg in kwargs.items()
            }
            yield cls(model, **i_kwargs)


# Attribute List ------------------------------------------------------------ #


class AttrIter(LocationSequence, Sequence):
    """Iterator over an attribute of objects in a sequence.

    Length, items access, and representation work like with a normal list.
    Calls are forwarded to each entry and return a list of return values.
    Boolean operators are applied to each entry and return a list of bools.
    Arithmetic operators are applied to each entry and return a new list.
    If applied to another `AttrList`, the first entry of the first list
    will be matched with the first entry of the second list, and so on.
    Else, the same value will be applied to each entry of the list.
    See :class:`AgentList` for examples.
    """

    def __init__(self, source, attr=None):
        """_summary_.

        Args:
            source (_type_): _description_
            attr (_type_, optional): _description_. Defaults to None.
        """
        self.source = source
        self.attr = attr

    def __repr__(self):
        """Return a REPL representation of LocationSequence.

        Returns:
            str: REPL representation.
        """
        return repr(list(self))

    @staticmethod
    def _iter_attr(a, s):
        for o in s:
            yield getattr(o, a)

    def __iter__(self):
        """Iterate through source list based on attribute."""
        if self.attr:
            return self._iter_attr(self.attr, self.source)
        else:
            return iter(self.source)

    def __len__(self):
        return len(self.source)

    def __getitem__(self, key):
        """Get item from source list."""
        if self.attr:
            return getattr(self.source[key], self.attr)
        else:
            return self.source[key]

    def __setitem__(self, key, value):
        """Set item to source list."""
        if self.attr:
            setattr(self.source[key], self.attr, value)
        else:
            self.source[key] = value

    def __call__(self, *args, **kwargs):
        return AttrIter([func_obj(*args, **kwargs) for func_obj in self])

    def __eq__(self, other):
        return [obj == other for obj in self]

    def __ne__(self, other):
        return [obj != other for obj in self]

    def __lt__(self, other):
        return [obj < other for obj in self]

    def __le__(self, other):
        return [obj <= other for obj in self]

    def __gt__(self, other):
        return [obj > other for obj in self]

    def __ge__(self, other):
        return [obj >= other for obj in self]

    def __add__(self, v):
        if isinstance(v, AttrIter):
            return AttrIter([x + y for x, y in zip(self, v)])
        else:
            return AttrIter([x + v for x in self])

    def __sub__(self, v):
        if isinstance(v, AttrIter):
            return AttrIter([x - y for x, y in zip(self, v)])
        else:
            return AttrIter([x - v for x in self])

    def __mul__(self, v):
        if isinstance(v, AttrIter):
            return AttrIter([x * y for x, y in zip(self, v)])
        else:
            return AttrIter([x * v for x in self])

    def __truediv__(self, v):
        if isinstance(v, AttrIter):
            return AttrIter([x / y for x, y in zip(self, v)])
        else:
            return AttrIter([x / v for x in self])

    def __iadd__(self, v):
        return self + v

    def __isub__(self, v):
        return self - v

    def __imul__(self, v):
        return self * v

    def __itruediv__(self, v):
        return self / v


class LocationList(LocationSequence, list):
    """List of agentpy objects.

    Attribute calls and assignments are applied to all agents
    and return an :class:`AttrIter` with the attributes of each agent.
    This also works for method calls, which returns a list of return values.
    Arithmetic operators can further be used to manipulate agent attributes,
    and boolean operators can be used to filter the list based on agents'
    attributes. Standard :class:`list` methods can also be used.

    Arguments:
        model (Model): The model instance.
        objs (int or Sequence, optional):
            An integer number of new objects to be created,
            or a sequence of existing objects (default empty).
        cls (type, optional): Class for the creation of new objects.
        **kwargs:
            Keyword arguments are forwarded
            to the constructor of the new objects.
            Keyword arguments with sequences of type :class:`AttrIter` will be
            broadcasted, meaning that the first value will be assigned
            to the first object, the second to the second, and so forth.
            Otherwise, the same value will be assigned to all objects.

    Examples:
        Prepare an :class:`AgentList` with three agents::

            >>> model = ap.Model()
            >>> agents = model.add_agents(3)
            >>> agents
            AgentList [3 agents]

        The assignment operator can be used to set a variable for each agent.
        When the variable is called, an :class:`AttrList` is returned::

            >>> agents.x = 1
            >>> agents.x
            AttrList of 'x': [1, 1, 1]

        One can also set different variables for each agent
        by passing another :class:`AttrList`::

            >>> agents.y = ap.AttrIter([1, 2, 3])
            >>> agents.y
            AttrList of 'y': [1, 2, 3]

        Arithmetic operators can be used in a similar way.
        If an :class:`AttrList` is passed, different values are used for
        each agent. Otherwise, the same value is used for all agents::

            >>> agents.x = agents.x + agents.y
            >>> agents.x
            AttrList of 'x': [2, 3, 4]

            >>> agents.x *= 2
            >>> agents.x
            AttrList of 'x': [4, 6, 8]

        Attributes of specific agents can be changed through setting items::

            >>> agents.x[2] = 10
            >>> agents.x
            AttrList of 'x': [4, 6, 10]

        Boolean operators can be used to select a subset of agents::

            >>> subset = agents(agents.x > 5)
            >>> subset
            AgentList [2 agents]

            >>> subset.x
            AttrList of attribute 'x': [6, 8]
    """

    def __init__(self, model, objs=(), cls=None, *args, **kwargs):
        if isinstance(objs, int):
            objs = self._obj_gen(model, objs, cls, *args, **kwargs)
        super().__init__(objs)
        super().__setattr__("model", model)
        super().__setattr__("ndim", 1)

    def __setattr__(self, name, value):
        if isinstance(value, AttrIter):
            # Apply each value to each agent
            for obj, v in zip(self, value):
                setattr(obj, name, v)
        else:
            # Apply single value to all agents
            for obj in self:
                setattr(obj, name, value)

    def __add__(self, other):
        agents = LocationList(self.model, self)
        agents.extend(other)
        return agents

    def select(self, selection):
        """Returns a new :class:`AgentList` based on `selection`.

        Arguments:
            selection (list of bool): List with same length as the agent list.
                Positions that return True will be selected.
        """
        return LocationList(self.model, [a for a, s in zip(self, selection) if s])

    def sort(self, var_key, reverse=False):
        """Sorts the list in-place, and returns self.

        Arguments:
            var_key (str): Attribute of the lists' objects, based on which
                the list will be sorted from lowest value to highest.
            reverse (bool, optional): Reverse sorting (default False).
        """
        super().sort(key=lambda x: x[var_key], reverse=reverse)
        return self

    def shuffle(self):
        """Shuffles the list in-place, and returns self."""
        self.model.random.shuffle(self)
        return self
